#include "StreamlineRenderer.h"
#include "VectorAttributes.h"
#include <GLUT/glut.h>									//GLUT library
#include <iostream>
#include <algorithm>
#include <set>


StreamlineRenderer::StreamlineRenderer(): max_length(1),min_length(0),step_size(0.2),thickness(1) 
{
}

StreamlineRenderer::~StreamlineRenderer()
{
	for(int i=0;i<streamlines.size();++i)						
		delete streamlines[i];
}


void StreamlineRenderer::computeVertex(const Volume<Point3d>& g,int i,int j,int k,float* vertex,float* result)
{														//Interpolate a 3D uniform volume from cell-data to vertex-data
														//For this, we use the classical scheme of averaging data
														//values (at each vertex) from the cells using that vertex

	int W = g.getWidth(), H = g.getHeight(), D = g.getDepth();
	
	Point3d sum;
	int	    count=0;

	for(int I=i;I<=i+1;++I)								//Iterate over all cells that share vertex (i,j,k)
	  if (I>=0 && I<W)									//Only consider cells that fall within the grid's bounds
	  for(int J=j;J<=j+1;++J)
		if (J>=0 && J<H)
		for(int K=k;K<=k+1;++K)
		{
			if (K<0 || K>=D) continue;
			const Point3d& v = g(I,J,K);				//Get cell-value of neighbor cell, and accumulate it in 'sum'
			sum += v;									//Also remember how many neighbor cells we found
			++count;
		}
	
	vertex[0] = i;										//Save vertex coordinate. By convention, this is equal
	vertex[1] = j;										//to the cell coordinate (i,j,k)
	vertex[2] = k;

	result[0] = sum.x/count;							//Compute vertex-value by averaging the
	result[1] = sum.y/count;							//values of all cells sharing that vertex
	result[2] = sum.z/count;
}



bool StreamlineRenderer::getC1Vector(const Volume<Point3d>& g,const float* position, float* value)
{
	int i = floor(position[0]), j = floor(position[1]), k = floor(position[2]);
														//1. Compute cell coordinates point is in.

	int I = g.getWidth(), J = g.getHeight(), K = g.getDepth();
	if (i<0 || i>=I || j<0 || j>=J || k<0 || k>=K) return false; 
														//2. Test if point is inside the volume. If not, we cannot interpolate there.

	float p[8][3],v[8][3];								// 'p' stores the coordinates of the eight vertices of cell (i,j,k)
														// 'v' stores the vector values at these vertices

	int c=0;											//3. Compute coordinates p[] and vector-values v[]
	for(int K=k-1;K<=k;++K)								//   at the eight vertices of cell (i,j,k)
	for(int J=j-1;J<=j;++J)								//   
	for(int I=i-1;I<=i;++I,++c)								
		computeVertex(g,I,J,K,p[c],v[c]);
					
	float tx = fabs(position[0]-p[0][0]);				//4. Parametric coordinates of 'position' within cell (i,j,k)
	float ty = fabs(position[1]-p[0][1]);				//   (should be all in [0,1])
	float tz = fabs(position[2]-p[0][2]);
		
	value[0]=value[1]=value[2]=0;						//Compute vector value at given point, using
														//trilinear interpolation of the vector values at the 8 cell vertices

	value[0] += (1-tx)*(1-ty)*(1-tz)*v[0][0];			
	value[1] += (1-tx)*(1-ty)*(1-tz)*v[0][1];
	value[2] += (1-tx)*(1-ty)*(1-tz)*v[0][2];

	value[0] += tx*(1-ty)*(1-tz)*v[1][0];
	value[1] += tx*(1-ty)*(1-tz)*v[1][1];
	value[2] += tx*(1-ty)*(1-tz)*v[1][2];

	value[0] += (1-tx)*ty*(1-tz)*v[2][0];
	value[1] += (1-tx)*ty*(1-tz)*v[2][1];
	value[2] += (1-tx)*ty*(1-tz)*v[2][2];

	value[0] += tx*ty*(1-tz)*v[3][0];
	value[1] += tx*ty*(1-tz)*v[3][1];
	value[2] += tx*ty*(1-tz)*v[3][2];

	value[0] += (1-tx)*(1-ty)*tz*v[4][0];			
	value[1] += (1-tx)*(1-ty)*tz*v[4][1];
	value[2] += (1-tx)*(1-ty)*tz*v[4][2];

	value[0] += tx*(1-ty)*tz*v[5][0];
	value[1] += tx*(1-ty)*tz*v[5][1];
	value[2] += tx*(1-ty)*tz*v[5][2];

	value[0] += (1-tx)*ty*tz*v[6][0];
	value[1] += (1-tx)*ty*tz*v[6][1];
	value[2] += (1-tx)*ty*tz*v[6][2];

	value[0] += tx*ty*tz*v[7][0];
	value[1] += tx*ty*tz*v[7][1];
	value[2] += tx*ty*tz*v[7][2];

	return true;										//Interpolation succeeded
}




float StreamlineRenderer::traceOneDirection(bool downstream,Volume<Point3d>& g,const Point3d& seed,StreamLine& streamline)
{                                                       //Trace one streamline in upstream or downstream direction	
    Point3d current = seed;
    float direction = (downstream)? 1:-1;
    
	float streamline_length = 0;						//Integrate the vector field for a given maximal length:
	for(int i=0;streamline_length<max_length;++i,streamline_length+=step_size)
	{
		float v[3];										//1. Get vector field at current position, using bilinear interpolation
		bool inside = getC1Vector(g,current.data,v);
		if (!inside) break;								//2. If current position is outside the grid, stop the tracing
        
		Point3d vec(v);
		float length  = vec.norm();
		
		StreamPoint p;									//3. Create a new streamline-point. Besides position, save the
		p.point     = current;							//normalized vector length, the distance along the streamline, and the vector value
		p.magnitude = (length-vector_min)/(vector_max-vector_min);
		p.position  = streamline_length;
		p.vector	= vec;
		
		streamline.points.push_back(p);					//4. Append point to streamline
        
		if (length<1.0e-5) break;						//5. If we reach an area of near-zero vector magnitude, stop tracing
        
		vec /= length;									//6. Make a new point by simple upstream/downstream Euler integration
                                                        //   Use here the normalized vector, for an even point distibution

        current += vec*direction*step_size;				//   along the streamline
	}
    
    return streamline_length; 
}



void StreamlineRenderer::trace(Volume<Point3d>& g,const Point3d& seed,StreamLine& streamline)				
{														//Trace a single streamline from 'seed' up- and downstream
    StreamLine upstream, downstream;

    float up_length   = traceOneDirection(true,g,seed,upstream);
    float down_length = traceOneDirection(false,g,seed,downstream);
    
    float total_length = up_length+down_length;			//Total length of upstream plus downstream lines
    
	streamline.points.reserve(upstream.points.size()+downstream.points.size());
														//Reserve space for final streamline (faster than resizing vectors)
	streamline.length = total_length;					//Remember the total length of the streamline								
	
    for(int i=int(upstream.points.size())-1;i>=0;--i)   //Add upstream points to final streamline, in inverse
    {                                                   //order of computation
        StreamPoint& p = upstream.points[i];
        float t = (up_length-p.position)/total_length;  //Arc-length parameter-value for p ([0,1] over entire streamline)
        p.position = t; 
        streamline.points.push_back(p);
    }

    for(int i=0;i<downstream.points.size();++i)         //Add downstream points to final streamline, in order
    {                                                   //of computation
        StreamPoint& p = downstream.points[i];
        float t = (p.position+up_length)/total_length;  //Arc-length parameter-value for p ([0,1] over entire streamline)
        p.position = t; 
        streamline.points.push_back(p);
    }
}



void StreamlineRenderer::computeStreamlines(Volume<Point3d>& g)	//Compute and cache streamlines for all seeds
{
	for(int i=0;i<streamlines.size();++i)				//1. Clear all cached streamlines since we make new ones		
		delete streamlines[i];
	streamlines.clear();								
	
	streamlines.reserve(seeds.size());

	vector_min = 1.0e6;									//2. Determine min and max of vector length,
	vector_max = -1;									//   used for later normalization
	for(int k=0;k<g.getDepth();++k)
	for(int j=0;j<g.getHeight();++j)
	for(int i=0;i<g.getWidth();++i)
	{
		const Point3d& v = g(i,j,k);
		float len = v.norm();
		vector_min = min(vector_min,len);
		vector_max = max(vector_max,len);
	}
	
	for(int i=0;i<seeds.size();++i)						//3. Trace a streamline for each seed
	{
		StreamLine* streamline = new StreamLine;
		const Point3d& seed = seeds[i];
		trace(g,seed,*streamline);						//3.1. Trace the streamline. Collect result in 'streamline'
		streamlines.push_back(streamline);				//3.2. Cache the computed streamline
	}
}



void StreamlineRenderer::draw(const Volume<float>& scalar)	 
{														//Draw computed streamlines
	glEnable(GL_BLEND);									
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);	//We want to use high-quality antialiased lines. For this, we need blending.
	glEnable(GL_LINE_SMOOTH);							//For this in turn, we need to disable writing to the Z buffer.
	glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_SMOOTH);
	glDepthMask(GL_FALSE);
	glLineWidth(thickness);

	for(int i=0;i<streamlines.size();++i)				//Draw all traced streamlines that are longer than min_length
	{
		const StreamLine& streamline = *streamlines[i];	
		if (streamline.length >= min_length)	
		   draw(streamline,scalar);
	}

	glDepthMask(GL_TRUE);
}


void StreamlineRenderer::draw(const StreamLine& streamline, const Volume<float>& scalar)
{														//Draw a streamline as polyline
    if (streamline.points.size()<2) return;				//We need at least 2 points to draw a streamline

	glBegin(GL_LINE_STRIP);
	for(int i=0;i<streamline.points.size();++i)
	{
		const StreamPoint& p = streamline.points[i];
				
		float R,G,B,A;									//Map streamline's direction to color, and the scalar field
		Point3d v = p.vector;							//to the streamline's transparency
		v.normalize();		
		R = fabs(v.x);
		G = fabs(v.y);
		B = fabs(v.z);
        A = scalar(int(p.point.x),int(p.point.y),int(p.point.z));
		
        glColor4f(R,G,B,A);								//Pass color and transparency to OpenGL
		glVertex3f(p.point.x,p.point.y,p.point.z);
	}
	glEnd();
}

